#! /usr/bin/env python

"""
Job control on LSF clusters (possibly connecting to the front-end via SSH).
"""

# Copyright (C) 2009-2016, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import next
__docformat__ = 'reStructuredText'


from collections import defaultdict
import datetime
import re
import time

from gc3libs import log, Run
import gc3libs.defaults
from gc3libs.backends import LRMS
import gc3libs.exceptions
from gc3libs.quantity import Duration, seconds, Memory, GB, MB, kB, bytes
import gc3libs.utils
from gc3libs.utils import sh_quote_safe_cmdline, sh_quote_unsafe_cmdline

from . import batch

# Examples of LSF commands output used to build this backend:
# $ bsub -W 00:10 -n 1 -R "rusage[mem=1800]" < script.sh
# Generic job.
# Job <473713> is submitted to queue <pub.1h>.
#
# where:
#   -W HH:MM is the wall-clock time
#   -n is the number of CPUs
#   -R sets resource limits
# note: script must be fed as STDIN
#
# [gloessa@brutus2 test_queue]$ bjobs 473713
# JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME   # noqa
# 473713  gloessa PEND  pub.1h     brutus2                 TM-T       Oct 19 17:10  # noqa
# noqa
# $ bjobs 473713                                                                    # noqa
# JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME   # noqa
# 473713  gloessa RUN   pub.1h     brutus2     a6128       TM-T       Oct 19 17:10  # noqa
# [gloessa@brutus2 test_queue]$ bjobs -l 473713                                     # noqa
# noqa
# Job <473713>, Job Name <TM-T>, User <gloessa>, Project <default>, Status <RUN>,   # noqa
# Queue <pub.1h>, Job Priority <50>, Command <#!/bin/sh;#;#   # noqa
# ;echo 'sequential test job';echo $OMP_NUM_THREADS;echo '$T   # noqa
# URBODIR: '$TURBODIR;sysname;which dscf;echo '$TMPDIR: '$TM   # noqa
# PDIR; echo 'parallel test job';export OMP_NUM_THREADS=2;ec   # noqa
# ho $OMP_NUM_THREADS;echo '$TURBODIR: '$TURBODIR;sysname;wh   # noqa
# ich dscf;echo '$TMPDIR: '$TMPDIR; sleep 300>, Share group    # noqa
# charged </lsf_cfour/gloessa>                                 # noqa
# Wed Oct 19 17:10:27: Submitted from host <brutus2>, CWD <$HOME/test_queue>, Out   # noqa
# put File <lsf.o%J>, Requested Resources <order[-r1m] span[   # noqa
# ptile=1] same[model] rusage[mem=1800,xs=1]>, Specified Hos   # noqa
# ts <thin+7>, <single+5>, <shared+3>, <parallel+1>;           # noqa
# noqa
# RUNLIMIT                                                                         # noqa
# 10.0 min of a6128                                                                # noqa
# Wed Oct 19 17:11:02: Started on <a6128>, Execution Home </cluster/home/chab/glo   # noqa
# essa>, Execution CWD </cluster/home/chab/gloessa/test_queu   # noqa
# e>;                                                          # noqa
# Wed Oct 19 17:12:10: Resource usage collected.                                    # noqa
# MEM: 5 Mbytes;  SWAP: 201 Mbytes;  NTHREAD: 5                # noqa
# PGID: 23177;  PIDs: 23177 23178 23182 23259                  # noqa
#
#
#  SCHEDULING PARAMETERS:
#            r15s   r1m  r15m   ut      pg    io   ls    it    tmp    swp   mem
#  loadSched   -     -     -     -       -     -    -     -     -      -     -
#  loadStop    -     -     -     -       -     -    -     -     -      -     -
#
#           scratch      xs       s       m       l      xl      sp
#  loadSched     -       -       -       -       -       -       -
#  loadStop      -       -       -       -       -       -       -
#
# $ bkill 473713
# Job <473713> is being terminated
#
# $ bjobs -W -w 473713
# JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME  PROJ_NAME CPU_USED MEM SWAP PIDS START_TIME FINISH_TIME  # noqa
# 473713  gloessa EXIT  pub.1h     brutus2     a6128       TM-T       10/19-17:10:27 default    000:00:00.12 5208   206312 23177,23178,23182,23259 10/19-17:11:02 10/19-17:14:49    # noqa
#
# STAT would be DONE if not killed
# $ bjobs 473713
# JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME  # noqa
# 473713  gloessa EXIT  pub.1h     brutus2     a6128       TM-T       Oct 19 17:10  # noqa
#
# $ bacct 473713
#
# Accounting information about jobs that are:
#   - submitted by all users.
#   - accounted on all projects.
#   - completed normally or exited
#   - executed on all hosts.
#   - submitted to all queues.
#   - accounted on all service classes.
# -----------------------------------------------------------------------------
#
# SUMMARY:      ( time unit: second )
#  Total number of done jobs:       0      Total number of exited jobs:     1
#  Total CPU time consumed:       0.1      Average CPU time consumed:     0.1
#  Maximum CPU time of a job:     0.1      Minimum CPU time of a job:     0.1
#  Total wait time in queues:    35.0
#  Average wait time in queue:   35.0
#  Maximum wait time in queue:   35.0      Minimum wait time in queue:   35.0
#  Average turnaround time:       262 (seconds/job)
#  Maximum turnaround time:       262      Minimum turnaround time:       262
#  Average hog factor of a job:  0.00 ( cpu time / turnaround time )
#  Maximum hog factor of a job:  0.00      Minimum hog factor of a job:  0.00
#
# $ lsinfo
# RESOURCE_NAME   TYPE   ORDER  DESCRIPTION
# r15s          Numeric   Inc   15-second CPU run queue length
# r1m           Numeric   Inc   1-minute CPU run queue length (alias: cpu)
# r15m          Numeric   Inc   15-minute CPU run queue length
# ut            Numeric   Inc   1-minute CPU utilization (0.0 to 1.0)
# pg            Numeric   Inc   Paging rate (pages/second)
# io            Numeric   Inc   Disk IO rate (Kbytes/second)
# ls            Numeric   Inc   Number of login sessions (alias: login)
# it            Numeric   Dec   Idle time (minutes) (alias: idle)
# tmp           Numeric   Dec   Disk space in /tmp (Mbytes)
# swp           Numeric   Dec   Available swap space (Mbytes) (alias: swap)
# mem           Numeric   Dec   Available memory (Mbytes)
# scratch       Numeric   Dec   Disk space available in /scratch in Mbytes
# xs            Numeric   Dec   number of slots available for express jobs
# s             Numeric   Dec   number of slots available for 1h jobs
# m             Numeric   Dec   number of slots available for 8h jobs
# l             Numeric   Dec   number of slots available for 36h jobs
# xl            Numeric   Dec   number of slots available for 7d jobs
# sp            Numeric   Dec   number of slots available for unlimited jobs
# ncpus         Numeric   Dec   Number of CPUs
# ndisks        Numeric   Dec   Number of local disks
# maxmem        Numeric   Dec   Maximum memory (Mbytes)
# maxswp        Numeric   Dec   Maximum swap space (Mbytes)
# maxtmp        Numeric   Dec   Maximum /tmp space (Mbytes)
# cpuf          Numeric   Dec   CPU factor
# rexpri        Numeric   N/A   Remote execution priority
# openclversion Numeric   Dec   e.g. 1.1
# nprocs        Numeric   Dec   Number of physical processors
# ncores        Numeric   Dec   Number of cores per physical processor
# nthreads      Numeric   Dec   Number of threads per processor core
# server        Boolean   N/A   LSF server host
# LSF_Base      Boolean   N/A   Base product
# lsf_base      Boolean   N/A   Base product
# LSF_Manager   Boolean   N/A   LSF Manager product
# lsf_manager   Boolean   N/A   LSF Manager product
# LSF_JobSchedu Boolean   N/A   JobScheduler product
# lsf_js        Boolean   N/A   JobScheduler product
# LSF_Make      Boolean   N/A   Make product
# lsf_make      Boolean   N/A   Make product
# LSF_Parallel  Boolean   N/A   Parallel product
# lsf_parallel  Boolean   N/A   Parallel product
# LSF_Analyzer  Boolean   N/A   Analyzer product
# lsf_analyzer  Boolean   N/A   Analyzer product
# mips          Boolean   N/A   MIPS architecture
# sparc         Boolean   N/A   SUN SPARC
# hpux          Boolean   N/A   HP-UX UNIX
# aix           Boolean   N/A   AIX UNIX
# irix          Boolean   N/A   IRIX UNIX
# rms           Boolean   N/A   RMS
# pset          Boolean   N/A   PSET
# dist          Boolean   N/A   DIST
# slurm         Boolean   N/A   SLURM
# cpuset        Boolean   N/A   CPUSET
# solaris       Boolean   N/A   SUN SOLARIS
# fs            Boolean   N/A   File server
# cs            Boolean   N/A   Compute server
# frame         Boolean   N/A   Hosts with FrameMaker licence
# bigmem        Boolean   N/A   Hosts with very big memory
# diskless      Boolean   N/A   Diskless hosts
# alpha         Boolean   N/A   DEC alpha
# linux         Boolean   N/A   LINUX UNIX
# nt            Boolean   N/A   Windows NT
# mpich_gm      Boolean   N/A   MPICH GM MPI
# lammpi        Boolean   N/A   LAM MPI
# mpichp4       Boolean   N/A   MPICH P4 MPI
# mvapich       Boolean   N/A   Infiniband MPI
# sca_mpimon    Boolean   N/A   SCALI MPI
# ibmmpi        Boolean   N/A   IBM POE MPI
# hpmpi         Boolean   N/A   HP MPI
# sgimpi        Boolean   N/A   SGI MPI
# intelmpi      Boolean   N/A   Intel MPI
# crayxt3       Boolean   N/A   Cray XT3 MPI
# crayx1        Boolean   N/A   Cray X1 MPI
# mpich_mx      Boolean   N/A   MPICH MX MPI
# mpichsharemem Boolean   N/A   MPICH Shared Memory
# mpich2        Boolean   N/A   MPICH2
# openmpi       Boolean   N/A   OPENMPI
# Platform_HPC  Boolean   N/A   platform hpc license
# platform_hpc  Boolean   N/A   platform hpc license
# fluent        Boolean   N/A   fluent availability
# ls_dyna       Boolean   N/A   ls_dyna availability
# nastran       Boolean   N/A   nastran availability
# pvm           Boolean   N/A   pvm availability
# openmp        Boolean   N/A   openmp availability
# ansys         Boolean   N/A   ansys availability
# blast         Boolean   N/A   blast availability
# gaussian      Boolean   N/A   gaussian availability
# lion          Boolean   N/A   lion availability
# scitegic      Boolean   N/A   scitegic availability
# schroedinger  Boolean   N/A   schroedinger availability
# hmmer         Boolean   N/A   hmmer availability
# ib            Boolean   N/A   InfiniBand QDR
# panfs         Boolean   N/A   Panasas file system
# c4            Boolean   N/A   C4 node with huge memory and scratch
# lustre        Boolean   N/A   Lustre file system
# define_ncpus_ Boolean   N/A   ncpus := procs
# define_ncpus_ Boolean   N/A   ncpus := cores
# define_ncpus_ Boolean   N/A   ncpus := threads
# mg            Boolean   N/A   Management hosts
# bluegene      Boolean   N/A   BLUEGENE
# cuda          Boolean   N/A   nodes supporting Nvidia CUDA
# cuda3         Boolean   N/A   nodes supporting Nvidia CUDA
# opencl        Boolean   N/A   nodes supporting OpenCL
# opencl3       Boolean   N/A   nodes supporting OpenCL
# nas           Boolean   N/A   all NAS shares are available
# imsb          Boolean   N/A   nodes supporting IMSB-user NAS shares
# perfctr       Boolean   N/A   nodes exclusive for perfctr, no panfs running
# type           String   N/A   Host type
# model          String   N/A   Host model
# status         String   N/A   Host status
# hname          String   N/A   Host name
# gpuvendor      String   N/A   e.g. Nvidia, AMD
# gputype        String   N/A   e.g. Tesla, Fermi, FireStream
# gpumodel       String   N/A   e.g. C1060, M2050, 9350
# cudaversion   Numeric   Dec   e.g. 3.2
# lic_fluent    Numeric   Dec   fluent base licenses
# lic_fluent_pa Numeric   Dec   fluent parallel licenses
# lic_cfx       Numeric   Dec   cfx base licenses
# lic_cfx_par   Numeric   Dec   cfx parallel licenses
# aa_r          Numeric   Dec   ansys academic research licenses
# aa_r_hpc      Numeric   Dec   ansys academic hpc licenses
# msi_tokenr    Numeric   Dec   msi tokens
# gpu           Numeric   Dec   nodes with 1,2,3... GPUs
# ddadb         Numeric   Dec   number of connections that ddadb.ethz.ch can handle    # noqa
#
# TYPE_NAME
# UNKNOWN_AUTO_DETECT
# DEFAULT
# DEFAULT
# CRAYJ
# CRAYC
# CRAYT
# CRAYSV1
# CRAYT3E
# CRAYX1
# DigitalUNIX
# ALPHA
# ALPHA5
# ALPHASC
# HPPA
# IBMAIX4
# IBMAIX532
# IBMAIX564
# LINUX
# LINUX2
# LINUXAXP
# LINUX86
# LINUXPPC
# LINUX64
# DLINUX
# DLINUX64
# DLINUXAXP
# SCYLD
# SLINUX
# SLINUX64
# NECSX4
# NECSX5
# NECSX6
# NECSX8
# NTX86
# NTX64
# NTIA64
# NTALPHA
# SGI5
# SGI6
# SUNSOL
# SOL732
# SOL64
# SGI64
# SGI65
# SGI658
# SOLX86
# SOLX8664
# HPPA11
# HPUXIA64
# MACOSX
# LNXS39032
# LNXS390X64
# LINUXPPC64
# BPROC
# BPROC4
# LINUX_ARM
# X86_64
# SX86_64
# IA64
# DIA64
# SIA64
#
# MODEL_NAME      CPU_FACTOR      ARCHITECTURE
# Opteron2216           8.00      x15_3604_AMDOpterontmProcessor2216
# Opteron2220           8.00      x15_3604_AMDOpterontmProcessor2220
# Opteron2435           8.20      x15_3604_AMDOpterontmProcessor2435
# Opteron8220           8.00      x15_3604_AMDOpterontmProcessor8220
# Opteron8380           7.50      x15_3604_AMDOpterontmProcessor8380
# Opteron8384           8.10      x15_3604_AMDOpterontmProcessor8384
# Opteron6174           7.00      x15_3604_AMDOpterontmProcessor6174
# [gloessa@brutus2 test_queue]$ lsid
# Platform LSF HPC 7 Update 6, Sep 04 2009
# Copyright 1992-2009 Platform Computing Corporation
#
# My cluster name is brutus
# My master name is hpcadm2


_bsub_jobid_re = re.compile(r'^Job <(?P<jobid>\d+)> is submitted', re.I)

# Job <850088>, Job Name <GdemoSimpleApp>, User <smaffiol>, Project <default>, Status <EXIT>, Queue <normal>, Command <./script.62b626ca7ad5acaf_01.sh>, Share group charged </smaffiol>  # noqa
# Wed Jul 11 14:11:10: Submitted from host <globus.vital-it.ch>, CWD <$HOME>, Output File (overwrite) <stdout.log>, Re-runnable, Login Shell </bin/sh>;  # noqa
# Wed Jul 11 14:11:47: Started on <cpt157>, Execution Home </home/smaffiol>, Execution CWD </home/smaffiol>;  # noqa
# Wed Jul 11 14:11:48: Exited with exit code 127. The CPU time used is 0.1 seconds.  # noqa
# Wed Jul 11 14:11:48: Completed <exit>.

_bjobs_long_re = re.compile(
    r'(?P<end_time>[a-zA-Z]+\s+[a-zA-Z]+\s+\d+\s+\d+:\d+:\d+):\s+'
    r'Exited with exit code (?P<exit_status>\d+)[^0-9]+'
    r'The CPU time used is (?P<used_cpu_time>[0-9\.]+)\s+'
)


class LsfLrms(batch.BatchSystem):

    """Job control on LSF clusters (possibly by connecting via SSH to a
    submit node).

    """

    _batchsys_name = 'LSF'

    def __init__(self, name,
                 # this are inherited from the base LRMS class
                 architecture, max_cores, max_cores_per_job,
                 max_memory_per_core, max_walltime,
                 auth,  # ignored if `transport` is 'local'
                 # these are inherited from `BatchSystem`
                 frontend, transport,
                 # these are specific to this backend
                 lsf_continuation_line_prefix_length=None,
                 # (Note that optional arguments to the `BatchSystem` class,
                 # e.g.:
                 #     keyfile=None, accounting_delay=15,
                 # are collected into `extra_args` and should not be explicitly
                 # spelled out in this signature.)
                 **extra_args):

        # init base class
        batch.BatchSystem.__init__(
            self, name,
            architecture, max_cores, max_cores_per_job,
            max_memory_per_core, max_walltime, auth,
            frontend, transport,
            **extra_args)

        self.bsub = self._get_command_argv('bsub')

        # LSF commands
        self._bacct = self._get_command('bacct', 'bjobs')
        self._bacct2 = self._get_command('bacct2', 'bacct')
        self._bjobs = self._get_command('bjobs')
        self._bkill = self._get_command('bkill')
        self._lshosts = self._get_command('lshosts')

        if lsf_continuation_line_prefix_length is not None:
            self._CONTINUATION_LINE_START = (' ' * lsf_continuation_line_prefix_length)
        else:
            self._CONTINUATION_LINE_START = None

    def _submit_command(self, app):
        # LSF's `bsub` allows one to submit scripts and binaries with
        # the same syntax, so we do not need to create an auxiliary
        # submission script and can just specify the command on the
        # command-line
        sub_argv, app_argv = app.bsub(self)
        prologue = self.get_prologue_script(app)
        epilogue = self.get_epilogue_script(app)
        if prologue or epilogue:
            return (sh_quote_safe_cmdline(sub_argv),
                    sh_quote_unsafe_cmdline(app_argv))
        else:
            return (sh_quote_unsafe_cmdline(sub_argv + app_argv), '')

    def _parse_submit_output(self, bsub_output):
        """Parse the ``bsub`` output for the local jobid."""
        return self.get_jobid_from_submit_output(bsub_output, _bsub_jobid_re)

    def _stat_command(self, job):
        return ("%s -l %s" % (self._bjobs, job.lrms_jobid))

    def _acct_command(self, job):
        return ("%s -l %s" % (self._bjobs, job.lrms_jobid))

    def _secondary_acct_command(self, job):
        return ("%s -l %s" % (self._bacct2, job.lrms_jobid))

    @staticmethod
    def _lsf_state_to_gc3pie_state(stat):
        log.debug("Translating LSF's `bjobs` status '%s' to"
                  " gc3libs.Run.State ...", stat)
        try:
            return {
                # LSF 'stat' mapping:
                'PEND': Run.State.SUBMITTED,
                'RUN': Run.State.RUNNING,
                'PSUSP': Run.State.STOPPED,
                'USUSP': Run.State.STOPPED,
                'SSUSP': Run.State.STOPPED,
                # DONE = successful termination
                'DONE': Run.State.TERMINATING,
                # EXIT = job was killed / exit forced
                'EXIT': Run.State.TERMINATING,
                # ZOMBI = job "killed" and unreachable
                'ZOMBI': Run.State.TERMINATING,
                'UNKWN': Run.State.UNKNOWN,
            }[stat]
        except KeyError:
            log.warning(
                "Unknown LSF job status '%s', returning `UNKNOWN`", stat)
            return Run.State.UNKNOWN

    _job_not_found_re = re.compile('Job <(?P<jobid>[0-9]+)> is not found')
    _status_re = re.compile(r'Status <(?P<state>[A-Z]+)>', re.M)
    _unsuccessful_exit_re = re.compile(
        r'Exited with exit code (?P<exit_status>[0-9]+).', re.M)
    _cpu_time_re = re.compile(
        r'The CPU time used is (?P<cputime>[0-9]+(\.[0-9]+)?) seconds', re.M)
    _mem_used_re = re.compile(
        r'MAX MEM:\s+(?P<mem_used>[0-9]+)\s+(?P<mem_unit>[a-zA-Z]+);', re.M)

    def _parse_stat_output(self, stdout, stderr):
        # LSF's `bjobs` can only report info for terminated jobs, if
        # they finished no longer than ``CLEAN_PERIOD`` seconds
        # before; for older jobs it just prints ``Job XXX is not
        # found`` to STDERR.  However, it does the same when passed a
        # non-existent job ID.  We cannot distinguish the two cases
        # here; let's just be optimistic and presume that if a job ID
        # is not found, it must have been terminated since (at least)
        # we have it in our records so it *was* submitted...  See
        # issue #513 for details.
        if self._job_not_found_re.match(stderr):
            return self._stat_result(Run.State.TERMINATING, None)

        # LSF `bjobs -l` uses a LDIF-style continuation lines, wherein
        # a line is truncated at 79 characters and continues upon the
        # next one; continuation lines start with a fixed amount of
        # whitespace.  However, the amount of whitespace varies with
        # LSF release and possibly other factors, so we need to guess
        # or have users configure it...
        if self._CONTINUATION_LINE_START is None:
            self._CONTINUATION_LINE_START = ' ' \
                * self._guess_continuation_line_prefix_len(stdout)

        # Join continuation lines, so that we can work on a single
        # block of text.
        lines = []
        for line in stdout.split('\n'):
            if len(line) == 0:
                continue
            if line.startswith(self._CONTINUATION_LINE_START):
                lines[-1] += line[len(self._CONTINUATION_LINE_START):]
            else:
                lines.append(line)

        # now rebuild stdout by joining the reconstructed lines
        stdout = '\n'.join(lines)

        state = Run.State.UNKNOWN
        termstatus = None

        # XXX: this only works if the current status is the first one
        # reported in STDOUT ...
        match = LsfLrms._status_re.search(stdout)
        if match:
            lsf_job_state = match.group('state')
            state = LsfLrms._lsf_state_to_gc3pie_state(lsf_job_state)
            if lsf_job_state == 'DONE':
                # DONE = success
                termstatus = (0, 0)
            elif lsf_job_state == 'EXIT':
                # EXIT = job exited with exit code != 0
                match = LsfLrms._unsuccessful_exit_re.search(stdout)
                if match:
                    exit_status = int(match.group('exit_status'))
                    termstatus = Run.shellexit_to_returncode(exit_status)

        return self._stat_result(state, termstatus)

    @staticmethod
    def _guess_continuation_line_prefix_len(stdout):
        """
        Guess the most likely length of the initial run of spaces in
        continuation lines in `bjobs` output.

        The euristics is rather crude: we count how many spaces are at
        the beginning of each line, and take the value with most
        occurrences.

        Since, in addition, LSF's `bjobs` output contains also
        scheduling parameters and other differently-formatted reports,
        we only consider lines that contain a ``<`` or a ``>``
        character to be valid continuation lines -- any other line is
        just discarded.

        This is necessary as the amount of whitespace at the beginning
        of lines seems to vary with LSF version and/or some other
        parameter.
        """
        # count occurrences of each prefix length
        occurrences = defaultdict(int)
        for line in stdout.split('\n'):
            if '<' not in line and '>' not in line:
                continue
            # FIXME: incorrect result if LSF mixes TABs and spaces
            ws_length = len(line) - len(line.lstrip())
            occurrences[ws_length] += 1
        # now find the length that has the max occurrences
        max_occurrences = max(occurrences.values())
        for length, count in list(occurrences.items()):
            if count == max_occurrences:
                return length

    # e.g., 'Mon Oct  8 12:04:56 2012'
    _TIMESTAMP_FMT_WITH_YEAR = '%a %b %d %H:%M:%S %Y'
    # e.g., 'Mon Oct  8 12:04:56'
    _TIMESTAMP_FMT_NO_YEAR = '%a %b %d %H:%M:%S'

    @staticmethod
    def _parse_timespec(ts):
        """Parse a timestamp as it appears in LSF bjobs/bacct logs."""
        # try "with year" format first, as it has all the info we need
        try:
            return datetime.datetime.strptime(
                ts, LsfLrms._TIMESTAMP_FMT_WITH_YEAR)
        except ValueError:
            pass  # ignore and try again without year
        try:
            # XXX: since we do not have a year, we resort to the
            # following heuristics: if the month in the timespec is
            # less than or equal to the current month, the timestamp
            # is for an event occurred during this year; if it's in a
            # month later than the current one, the timestamp refers
            # to an event occurred in the *past* year.
            today = datetime.date.today()
            # XXX: datetime.strptime() only available starting Py 2.5
            tm = time.strptime(ts, LsfLrms._TIMESTAMP_FMT_NO_YEAR)
            if tm[1] <= today.month:
                return datetime.datetime(today.year, *(tm[1:6]))
            else:
                return datetime.datetime(today.year - 1, *(tm[1:6]))
        except ValueError as err:
            gc3libs.log.error(
                "Cannot parse '%s' as an LSF timestamp: %s: %s",
                ts, err.__class__.__name__, err)
            raise

    @staticmethod
    def _parse_memspec(m):
        unit = m[-1]
        if unit == 'G':
            return Memory(int(m[:-1]), unit=GB)
        elif unit == 'M':
            return Memory(int(m[:-1]), unit=MB)
        elif unit in ['K', 'k']:  # XXX: not sure which one is used
            return Memory(int(m[:-1]), unit=kB)
        else:
            # XXX: what does LSF use as a default?
            return Memory(int(m), unit=bytes)

    def _parse_acct_output(self, stdout, stderr):
        # FIXME: this is an ugly fix, but we have issues with bacct
        # on some LSF installation being veeeeery slow, so we have to
        # try and use `bjobs` whenever possible, and fall back to
        # bacct if bjobs does not work.
        #
        # However, since the user could update the configuration file
        # and put `bacct = bacct`, we also have to ensure that we are
        # calling the correct function to parse the output of the acct
        # command.
        if 'bacct' in self._bacct:
            parser = self.__parse_acct_output_w_bacct
        elif 'bjobs' in self._bacct:
            parser = self.__parse_acct_output_w_bjobs
        else:
            log.warning(
                "Unknown accounting command `%s`."
                " Assuming its output is compatible"
                " with `bacct`", self._bacct)
            parser = self.__parse_acct_output_w_bacct
        return parser(stdout)

    _parse_secondary_acct_output = _parse_acct_output

    @staticmethod
    def __parse_acct_output_w_bjobs(stdout):
        acctinfo = {}

        # Try to parse used cputime
        match = LsfLrms._cpu_time_re.search(stdout)
        if match:
            cpu_time = match.group('cputime')
            acctinfo['used_cpu_time'] = Duration(float(cpu_time), unit=seconds)

        # Parse memory usage
        match = LsfLrms._mem_used_re.search(stdout)
        if match:
            mem_used = match.group('mem_used')
            # mem_unit should always be Mbytes
            acctinfo['max_used_memory'] = Memory(float(mem_used), unit=MB)

        # Find submission time and completion time
        for line in stdout.split('\n'):
            match = LsfLrms._EVENT_RE.match(line)
            if match:
                timestamp = line.split(': ')[0]
                event = match.group('event')
                if event == 'Submitted':
                    acctinfo['lsf_submission_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                elif event in ['Dispatched', 'Started']:
                    acctinfo['lsf_start_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                elif event in ['Completed', 'Done successfully']:
                    acctinfo['lsf_completion_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                continue
        if 'lsf_completion_time' in acctinfo and 'lsf_start_time' in acctinfo:
            acctinfo['duration'] = Duration(
                acctinfo['lsf_completion_time'] - acctinfo['lsf_start_time'])
        else:
            # XXX: what should we use for jobs that did not run at all?
            acctinfo['duration'] = Duration(0, unit=seconds)

        return acctinfo


    _RESOURCE_USAGE_RE = re.compile(r'^\s+ CPU_T \s+ '
                                    r'WAIT \s+ '
                                    r'TURNAROUND \s+ '
                                    r'STATUS \s+ '
                                    r'HOG_FACTOR \s+ '
                                    r'MEM \s+ '
                                    r'SWAP', re.X)
    _EVENT_RE = re.compile(
        r'^(Mon|Tue|Wed|Thu|Fri|Sat|Sun)'
        r' \s+ (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)'
        r' \s+ [0-9]+ \s+ [0-9:]+:'
        r' \s+ (?P<event>Submitted|Dispatched|Started|Completed|'
        r'Done\ successfully)', re.X)

    @staticmethod
    def __parse_acct_output_w_bacct(stdout):
        acctinfo = {}
        lines = iter(stdout.split('\n'))  # need to lookup next line in the loop
        for line in lines:
            match = LsfLrms._EVENT_RE.match(line)
            if match:
                timestamp = line.split(': ')[0]
                event = match.group('event')
                if event == 'Submitted':
                    acctinfo['lsf_submission_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                elif event == 'Dispatched':
                    acctinfo['lsf_start_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                elif event == 'Completed':
                    acctinfo['lsf_completion_time'] = \
                        LsfLrms._parse_timespec(timestamp)
                continue
            match = LsfLrms._RESOURCE_USAGE_RE.match(line)
            if match:
                # actual resource usage is on next line
                rusage = next(lines)
                cpu_t, wait, turnaround, status, hog_factor, mem, swap = \
                    rusage.split()
                # common backend attrs (see Issue 78)
                if 'lsf_completion_time' in acctinfo and 'lsf_start_time' in acctinfo:
                    acctinfo['duration'] = Duration(
                        acctinfo['lsf_completion_time'] - acctinfo['lsf_start_time'])
                else:
                    # XXX: what should we use for jobs that did not run at all?
                    acctinfo['duration'] = Duration(0, unit=seconds)
                acctinfo['used_cpu_time'] = Duration(float(cpu_t), unit=seconds)
                acctinfo['max_used_memory'] = LsfLrms._parse_memspec(mem)\
                    + LsfLrms._parse_memspec(swap)
                # the resource usage line is the last interesting line
                break
        return acctinfo

    def _cancel_command(self, jobid):
        return ("%s %s" % (self._bkill, jobid))

    @gc3libs.utils.cache_for(gc3libs.defaults.LSF_CACHE_TIME)
    @LRMS.authenticated
    def get_resource_status(self):
        """
        Get dynamic information out of the LSF subsystem.

        return self

        dynamic information required (at least those):
        total_queued
        free_slots
        user_running
        user_queued
        """

        try:
            self.transport.connect()

            # Run lhosts to get the list of available nodes and their
            # related number of cores
            # used to compute self.total_slots
            # lhost output format:
            # ($nodeid,$OStype,$model,$cpuf,$ncpus,$maxmem,$maxswp)
            _command = ('%s -w' % self._lshosts)
            exit_code, stdout, stderr = self.transport.execute_command(
                _command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "LSF backend failed executing '%s':"
                    "exit code: %d; stdout: '%s'; stderr: '%s'." %
                    (_command, exit_code, stdout, stderr))

            if stdout:
                lhosts_output = stdout.strip().split('\n')
                # Remove Header
                lhosts_output.pop(0)
            else:
                lhosts_output = []

            # compute self.total_slots
            self.max_cores = 0
            for line in lhosts_output:
                # HOST_NAME      type    model  cpuf ncpus maxmem maxswp server RESOURCES  # noqa
                (hostname, h_type, h_model, h_cpuf, h_ncpus) = \
                    line.strip().split()[0:5]
                try:
                    self.max_cores += int(h_ncpus)
                except ValueError:
                    # h_ncpus == '-'
                    pass

            # Run `bjobs -u all -w` to get information about the jobs
            # for a given user used to compute `running_jobs`,
            # `self.queued`, `self.user_run` and `self.user_queued`.
            #
            # bjobs output format:
            # JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME  # noqa
            _command = ('%s -u all -w' % self._bjobs)
            log.debug("Runing `%s`... ", _command)
            exit_code, stdout, stderr = \
                self.transport.execute_command(_command)
            if exit_code != 0:
                # cannot continue
                raise gc3libs.exceptions.LRMSError(
                    "LSF backend failed executing '%s':"
                    "exit code: %d; stdout: '%s'; stderr: '%s'." %
                    (_command, exit_code, stdout, stderr))

            if stdout:
                bjobs_output = stdout.strip().split('\n')
                # Remove Header
                bjobs_output.pop(0)
            else:
                bjobs_output = []

            # user runing/queued
            used_cores = 0
            self.queued = 0
            self.user_queued = 0
            self.user_run = 0

            queued_statuses = ['PEND', 'PSUSP', 'USUSP',
                               'SSUSP', 'WAIT', 'ZOMBI']
            for line in bjobs_output:
                # JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME  # noqa
                (jobid, user, stat, queue, from_h, exec_h) = \
                    line.strip().split()[0:6]
                # to compute the number of cores allocated per each job
                # we use the output format of EXEC_HOST field
                # e.g.: 1*cpt178:2*cpt151
                for node in exec_h.split(':'):
                    try:
                        # multi core
                        (cores, n_name) = node.split('*')
                    except ValueError:
                        # single core
                        cores = 1
                try:
                    cores = int(cores)
                except ValueError:
                    # core == '-'
                    pass
                used_cores += cores

                if stat in queued_statuses:
                    self.queued += 1
                if user == self._username:
                    if stat in queued_statuses:
                        self.user_queued += 1
                    else:
                        self.user_run += 1

            self.free_slots = self.max_cores - used_cores

            return self

        except Exception as ex:
            # self.transport.close()
            log.error("Error querying remote LRMS, see debug log for details.")
            log.debug("Error querying LRMS: %s: %s",
                      ex.__class__.__name__, str(ex))
            raise


# main: run tests
if "__main__" == __name__:
    import doctest
    doctest.testmod(name="lsf",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
"""
"""
# Copyright (C) 2011, 2012, 2016,  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from builtins import object
__docformat__ = 'reStructuredText'

import datetime
import os
import shutil
import tempfile

import pytest

import gc3libs
import gc3libs.core
import gc3libs.config
from gc3libs.quantity import kB, seconds, minutes

from faketransport import FakeTransport

files_to_remove = []


def correct_submit(jobid=123):
    out = """%s
    """ % jobid
    return (0, out, "")


def tracejob_notfound(jobid=123):
    err = """/var/spool/torque/server_priv/accounting/20120309: \
Permission denied
/var/spool/torque/mom_logs/20120309: No such file or directory
/var/spool/torque/sched_logs/20120309: No such file or directory
"""

    return (0, "", err)


def correct_qstat_queued(jobid=123):
    out = """%s                 antani     amessina                0 Q short
""" % jobid
    return (0, out, "")


def correct_tracejob_queued(jobid=123):
    out = """Job: %s

03/14/2012 18:42:04  S    enqueuing into short, state 1 hop 1
03/14/2012 18:42:04  S    Job Queued at request of amessina@argo.ictp.it, \
owner = amessina@argo.ictp.it, job name = STDIN, queue = short
""" % jobid
    err = """/var/spool/torque/server_priv/accounting/20120314: \
Permission denied
/var/spool/torque/mom_logs/20120314: No such file or directory
/var/spool/torque/sched_logs/20120314: No such file or directory
"""
    return (0, out, err)


def correct_qstat_running(jobid=123):
    out = """%s                 cam_icbc         idiallo         \
01:01:18 R short
""" % jobid
    return (0, out, "")


def correct_tracejob_running(jobid=123):
    out = """Job: %s

03/14/2012 18:42:04  S    enqueuing into short, state 1 hop 1
03/14/2012 18:42:04  S    Job Queued at request of amessina@argo.ictp.it, \
owner = amessina@argo.ictp.it, job name = STDIN, queue = short
03/14/2012 18:45:25  S    Job Run at request of root@argo.ictp.it
03/14/2012 18:45:25  S    Not sending email: User does not want mail \
of this type.
""" % jobid
    err = """/var/spool/torque/server_priv/accounting/20120314: \
Permission denied
/var/spool/torque/mom_logs/20120314: No such file or directory
/var/spool/torque/sched_logs/20120314: No such file or directory
"""
    return (0, out, err)


def qdel_notfound(jobid=123):
    return (153, "", """qdel: Unknown Job Id %s""" % jobid)


def qstat_notfound(jobid=123):
    err = """qstat: Unknown Job Id %s
""" % jobid
    return (153, "", err)


def correct_tracejob_done(jobid=123):
    out = """
Job: %s

03/09/2012 09:31:53  S    enqueuing into short, state 1 hop 1
03/09/2012 09:31:53  S    Job Queued at request of amessina@argo.ictp.it, \
owner = amessina@argo.ictp.it, job name = DemoPBSApp, queue = short
03/09/2012 09:32:03  S    Job Run at request of root@argo.ictp.it
03/09/2012 09:32:03  S    Not sending email: User does not want mail \
of this type.
03/09/2012 09:34:08  S    Not sending email: User does not want mail \
of this type.
03/09/2012 09:34:08  S    Exit_status=0 resources_used.cput=00:00:00 \
resources_used.mem=2364kb resources_used.vmem=190944kb \
resources_used.walltime=00:02:05
03/09/2012 09:34:08  S    dequeuing from short, state COMPLETE
""" % jobid

    err = """/var/spool/torque/server_priv/accounting/20120309: \
Permission denied
/var/spool/torque/mom_logs/20120309: No such file or directory
/var/spool/torque/sched_logs/20120309: No such file or directory
"""

    return (0, out, err)


def qsub_failed_resources():
    out = ""
    err = """qsub: Job exceeds queue resource limits MSG=cannot locate \
feasible nodes
"""
    return (190, out, err)


def qsub_failed_acl():
    out = ""
    err = """qsub: Unauthorized Request  MSG=user ACL rejected the \
submitting user: user amessina@argo.ictp.it, queue cm1
"""
    return (159, out, err)


def qdel_success():
    return (0, "", "")


def qdel_failed_acl(jobid=123):
    err = """qdel: Unauthorized Request  MSG=operation not permitted %s
""" % jobid
    out = ""
    return (159, out, err)

# import gc3libs.Run.State as State
State = gc3libs.Run.State


class FakeApp(gc3libs.Application):

    def __init__(self):
        gc3libs.Application.__init__(
            self,
            arguments=['/bin/hostname'],  # mandatory
            inputs=[],                    # mandatory
            outputs=[],                   # mandatory
            output_dir="./fakedir",       # mandatory
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1,)


class TestBackendPbs(object):

    CONF = """
[resource/example]
type=pbs
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64
queue=testing
enabled=True

[auth/ssh]
type=ssh
username=NONEXISTENT
"""
    @pytest.fixture(autouse=True)
    def setUp(self):
        (fd, self.tmpfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendPbs.CONF)
        f.close()

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(self.tmpfile)

        self.core = gc3libs.core.Core(self.cfg)

        self.backend = self.core.get_backend('example')
        self.backend.transport = FakeTransport()
        self.transport = self.backend.transport
        # Basic responses
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['tracejob'] = tracejob_notfound()
        self.transport.expected_answer['qdel'] = qdel_notfound()

        yield

        os.remove(self.tmpfile)

    def test_submission_failed(self):
        app = FakeApp()

        # Submission failed (unable to find resources):
        self.transport.expected_answer['qsub'] = qsub_failed_resources()
        try:
            self.core.submit(app)
        except Exception as e:
            assert isinstance(e, gc3libs.exceptions.LRMSError)
        assert app.execution.state == State.NEW

        # Submission failed (unauthrozed user):
        self.transport.expected_answer['qsub'] = qsub_failed_acl()
        try:
            self.core.submit(app)
        except Exception as e:
            assert isinstance(e, gc3libs.exceptions.LRMSError)
        assert app.execution.state == State.NEW

    def test_pbs_basic_workflow(self):
        app = FakeApp()

        # Succesful submission:
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

        # Update state. We would expect the job to be SUBMITTED
        self.transport.expected_answer['qstat'] = correct_qstat_queued()
        self.transport.expected_answer['tracejob'] = correct_tracejob_queued()
        self.core.update_job_state(app)
        assert app.execution.state == State.SUBMITTED

        # Update state. We would expect the job to be RUNNING
        self.transport.expected_answer['qstat'] = correct_qstat_running()
        self.transport.expected_answer['tracejob'] = correct_tracejob_running()
        self.core.update_job_state(app)
        assert app.execution.state == State.RUNNING

        # Job done. qstat doesn't find it, tracejob should.
        self.transport.expected_answer['qstat'] = qstat_notfound()
        self.transport.expected_answer['tracejob'] = correct_tracejob_done()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

    def test_tracejob_parsing(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        self.transport.expected_answer['tracejob'] = correct_tracejob_done()
        self.core.update_job_state(app)
        assert app.execution.state == State.TERMINATING

        job = app.execution
        # common job reporting values (see Issue 78)
        assert job.exitcode == 0
        assert job.duration == 2 * minutes + 5 * seconds
        assert job.max_used_memory == 190944 * kB
        assert job.used_cpu_time == 0 * seconds
        # PBS-specific values
        assert job.pbs_queue == 'short'
        assert job.pbs_jobname == 'DemoPBSApp'
        assert job.pbs_max_used_ram == 2364 * kB
        assert (job.pbs_submission_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=31,
                                       second=53))
        assert (job.pbs_running_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=32,
                                       second=3))
        assert (job.pbs_end_time ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=34,
                                       second=8))

    def test_delete_job(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)

        self.transport.expected_answer['qdel'] = qdel_success()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED

    def test_delete_already_deleted_job(self):
        app = FakeApp()
        self.transport.expected_answer['qsub'] = correct_submit()
        self.core.submit(app)
        assert app.execution.state == State.SUBMITTED

        self.transport.expected_answer['qdel'] = qdel_failed_acl()
        self.core.kill(app)
        assert app.execution.state == State.TERMINATED

    def test_parse_acct_output(self):
        rc, stdout, stderr = correct_tracejob_done()
        status = self.backend._parse_acct_output(stdout, stderr)
        # common job reporting values (see Issue 78)
        assert status['termstatus'] == (0, 0)
        assert status['duration'] == 2 * minutes + 5 * seconds
        assert status['max_used_memory'] == 190944 * kB
        assert status['used_cpu_time'] == 0 * seconds
        # PBS-specific values
        assert status['pbs_queue'] == 'short'
        assert status['pbs_jobname'] == 'DemoPBSApp'
        assert status['pbs_max_used_ram'] == 2364 * kB
        assert (status['pbs_submission_time'] ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=31,
                                       second=53))
        assert (status['pbs_running_time'] ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=32,
                                       second=3))
        assert (status['pbs_end_time'] ==
                     datetime.datetime(year=2012,
                                       month=3,
                                       day=9,
                                       hour=9,
                                       minute=34,
                                       second=8))


def tearDownModule():
    for fname in files_to_remove:
        if os.path.isdir(fname):
            shutil.rmtree(fname)
        else:
            os.remove(fname)


def test_get_command():
    (fd, tmpfile) = tempfile.mkstemp()
    files_to_remove.append(tmpfile)
    f = os.fdopen(fd, 'w+')
    f.write("""
[auth/ssh]
type=ssh
username=NONEXISTENT

[resource/example]
# mandatory stuff
type=pbs
auth=ssh
transport=ssh
frontend=example.org
max_cores_per_job=128
max_memory_per_core=2
max_walltime=2
max_cores=80
architecture=x86_64

# alternate command paths
qsub = /usr/local/bin/qsub -q testing
qstat = /usr/local/bin/qstat
qdel = /usr/local/bin/qdel # comments are ignored!
tracejob = /usr/local/sbin/tracejob
""")
    f.close()

    cfg = gc3libs.config.Configuration()
    cfg.merge_file(tmpfile)
    b = cfg.make_resources()['example']

    assert b.qsub == ['/usr/local/bin/qsub', '-q', 'testing']

    assert b._qstat == '/usr/local/bin/qstat'
    assert b._qdel == '/usr/local/bin/qdel'
    assert b._tracejob == '/usr/local/sbin/tracejob'


if __name__ == "__main__":
    import pytest
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gchelsa.py -- Front-end script for running chelsa application
#   driven by a custom made R script.
#
#   Copyright (C) 2016, 2019  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gchelsa.py --help`` for program usage
instructions.

Input argument consists of:
- Month range: integer range (e.g. 1:430)
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-02-109:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gchelsa
    gchelsa.GchelsaScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(62000,MB)

DEFAULT_REMOTE_TEMP="/tmp"
DEFAULT_REMOTE_INPUT="/data/input"
DEFAULT_REMOTE_OUTPUT="/data/output"
DEFAULT_REMOTE_OUTPUT_FILE="./output.txt"
DEFAULT_CHELSA_SCRIPT="/INPUT/CHELSA_LATLONG_HPC2.R"

## custom application class
class GchelsaApplication(Application):
    """
    """
    application_name = 'gchelsa'

    def __init__(self, month, **extra_args):
        """
        Remote command execution: R --vanilla --args "c(START,END) /tmp/ /mnt/s3it/out/ /mnt/" </mnt/CHELSA_LATLONG_HPC2.R >outfile.tx
        """
        inputs = dict()
        outputs = dict()

        if 'Rscript' in extra_args:
            chelsa_script = os.path.basename(extra_args['Rscript'])
            inputs[extra_args['Rscript']] = chelsa_script
        else:
            chelsa_script = DEFAULT_CHELSA_SCRIPT

        execution_script = """
#!/bin/sh

tempdir=`mktemp -d -p %s`
ulimit -a
# execute command
R --vanilla --args "gTIME=c(%s,%s) $tempdir/ %s %s" <%s>%s
RET=$?

echo Program terminated with exit code $RET
exit $RET
        """ % (extra_args['temp_data'],
               month,
               month,
               extra_args['output_data'],
               extra_args['input_data'],
               chelsa_script,
               DEFAULT_REMOTE_OUTPUT_FILE)

        try:
            # create script file
            (handle, self.tmp_filename) = tempfile.mkstemp(prefix='gchelsa-',
                                                           suffix=extra_args['jobname'])

            # XXX: use NamedTemporaryFile instead with 'delete' = False
            fd = open(self.tmp_filename,'w')
            fd.write(execution_script)
            fd.close()
            os.chmod(fd.name,0o777)
        except Exception, ex:
            gc3libs.log.debug("Error creating execution script" +
                              "Error type: %s." % type(ex) +
                              "Message: %s"  %ex.message)
            raise

        inputs[fd.name] = './gchelsa_wrapper.sh'

        extra_args['requested_memory'] = DEFAULT_MEMORY

        Application.__init__(
            self,
            arguments = ['./gchelsa_wrapper.sh'],
            inputs = inputs,
            outputs = [DEFAULT_REMOTE_OUTPUT_FILE],
            stdout = 'gchelsa.log',
            join=True,
            executables = ['gchelsa_wrapper.sh'],
            **extra_args)

class GchelsaScript(SessionBasedScript):
    """

    The ``gchelsa`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gchelsa`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GchelsaApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GchelsaApplication,
            )

    def setup_args(self):

        self.add_param('range', type=str,
                       help="Months range. "
                       "Format: [int]|[int]:[int]. E.g 1:432|3")

    def setup_options(self):
        self.add_param("-R", "--Rscript", metavar="STRING", type=str,
                       dest="Rscript",
                       default=None,
                       help="Location of master R script to drive the "
                       " execution of chelsa. Default: %(default)s")

        self.add_param("-I", "--input", metavar="PATH", type=str,
                       dest="input_data",
                       default=DEFAULT_REMOTE_INPUT,
                       help="Location of the input data folder. "
                       "Default: %(default)s")

        self.add_param("-O", "--output", metavar="PATH", type=str,
                       dest="output_data",
                       default=DEFAULT_REMOTE_OUTPUT,
                       help="Location of the output data folder. "
                       "Default: %(default)s")

        self.add_param("-T", "--temp", metavar="PATH", type=str,
                       dest="temp_data",
                       default=DEFAULT_REMOTE_TEMP,
                       help="Location of the temp data folder. "
                       "Default: %(default)s")

    def parse_args(self):
        try:
            if self.params.Rscript:
                assert os.path.isfile(self.params.Rscript), \
                    "R script file %s not found" % self.params.Rscript

            # Validate month range
            try:
                self.input_range = [ int(mrange) for mrange in \
                                     self.params.range.split(":") \
                                     if int(mrange) ]

                if len(self.input_range) == 1:
                    # Defined only single month
                    gc3libs.log.info("Defined single month to process: '%d'",
                                     self.input_range[0])
                elif len(self.input_range) == 2:
                    # Defined a range
                    self.input_range = range(self.input_range[0],
                                             self.input_range[1]+1)
                else:
                    # Anything else should fail
                    raise ValueError("No valid input range. "
                                         "Format: [int]|[int]:[int]. E.g 1:432|3")

            except ValueError as ex:
                gc3libs.log.debug(ex.message)
                raise AttributeError("No valid input range. "
                                     "Format: [int]|[int]:[int]. E.g 1:432|3")

        except AssertionError as ex:
            raise OSError(ex.message)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GchelsaApplication
        """
        tasks = []

        for month in self.input_range:

            # filename example: 0103645_anat.nii.gz
            # extract root folder name to be used as jobname
            extra_args = extra.copy()
            extra_args['jobname'] = 'chelsa-%s' % str(month)

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME',
                                                                        'run_%s' % month)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION',
                                                                        'run_%s' % month)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE',
                                                                        'run_%s' % month)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME',
                                                                        'run_%s' % month)

            if self.params.Rscript:
                extra_args['Rscript'] = os.path.abspath(self.params.Rscript)

            extra_args['input_data'] = self.params.input_data
            extra_args['output_data'] = self.params.output_data
            extra_args['temp_data'] = self.params.temp_data

            tasks.append(GchelsaApplication(
                month,
                **extra_args))

        return tasks



#! /usr/bin/env python

"""
Support and expansion of programmatic templates.

The module `gc3libs.template` allows creation of textual templates
with a simple object-oriented programming interface: given a string
with a list of substitutions (using the syntax of Python's standard
`substitute` module), a set of replacements can be specified, and the
`gc3libs.template.expansions` function will generate all possible
texts coming from the same template.  Templates can be nested, and
expansions generated recursviely.
"""

# Copyright (C) 2009-2012, 2014, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'


import string
import itertools

SetProductIterator = itertools.product


class Template(object):

    """
    A template object is a pair `(obj, keywords)`.  Methods are
    provided to substitute the keyword values into `obj`, and to
    iterate over expansions of the given keywords (optionally
    filtering the allowed combination of keyword values).

    Second optional argument `validator` must be a function that
    accepts a set of keyword arguments, and returns `True` if the
    keyword combination is valid (can be expanded/substituted back
    into the template) or `False` if it should be discarded.
    The default validator passes any combination of keywords/values.
    """

    def __init__(self, template,
                 validator=(lambda kws: True), **extra_args):
        self._template = template
        self._keywords = extra_args
        self._valid = validator

    def substitute(self, **extra_args):
        """
        Return result of interpolating the value of keywords into the
        template.  Keyword arguments `extra_args` can be used to override
        keyword values passed to the constructor.

        If the templated object provides a `substitute` method, then
        return the result of invoking it with the template keywords as
        keyword arguments.  Otherwise, return the result of applying
        Python standard library's `string.Template.safe_substitute()`
        on the string representation of the templated object.

        Raise `ValueError` if the set of keywords/values is not valid
        according to the validator specified in the constructor.
        """
        keywords = self._keywords.copy()
        keywords.update(extra_args)
        if self._valid(keywords):
            try:
                return self._template.substitute(**keywords)
            except AttributeError:
                return string.Template(
                    str(self._template)).safe_substitute(keywords)
        else:
            raise ValueError("Invalid substitution values in template.")

    def __eq__(self, other):
        return (
            self._template == other._template
            and self._keywords == other._keywords
            and self._valid == other._valid
        )

    def __str__(self):
        """Alias for `Template.substitute`."""
        return self.substitute()

    def __repr__(self):
        """
        Return a string representation such that `x == eval(repr(x))`.
        """
        return ''.join(["Template(",
                             ', '.join([repr(self._template)] +
                                      [("%s=%s" % (k, v))
                                       for k, v in list(self._keywords.items())]),
                             ')'])

    def expansions(self, **keywords):
        """
        Iterate over all valid expansions of the templated object
        *and* the template keywords.  Returned items are `Template`
        instances constucted with the expanded template object and a
        valid combination of keyword values.
        """
        keywords_ = self._keywords.copy()
        keywords_.update(keywords)
        for kws in expansions(keywords_):
            for item in expansions(self._template, **kws):
                # propagate expanded keywords upwards;
                # this allows container templates to
                # check validity based on keywords expanded
                # on contained templates as well.
                new_kws = kws.copy()
                for v in list(kws.values()):
                    if isinstance(v, Template):
                        new_kws.update(v._keywords)
                if self._valid(new_kws):
                    yield Template(item, self._valid, **new_kws)


def expansions(obj, **extra_args):
    """
    Iterate over all expansions of a given object, recursively
    expanding all templates found.  How the expansions are actually
    computed, depends on the type of object being passed in the first
    argument `obj`:

    * If `obj` is a `list`, iterate over expansions of items in `obj`.
      (In particular, this flattens out nested lists.)

      Example::

        >>> L = [0, [2, 3]]
        >>> list(expansions(L))
        [0, 2, 3]

    * If `obj` is a dictionary, return dictionary formed by all
      combinations of a key `k` in `obj` with an expansion of the
      corresponding value `obj[k]`.  Expansions are computed by
      recursively calling `expansions(obj[k], **extra_args)`.

      Example::

        >>> D = {'a':1, 'b':[2,3]}
        >>> E = list(expansions(D))
        >>> len(E)
        2
        >>> {'a': 1, 'b': 2} in E
        True
        >>> {'a': 1, 'b': 3} in E
        True

    * If `obj` is a `tuple`, iterate over all tuples formed by the
      expansion of every item in `obj`.  (Each item `t[i]` is expanded
      by calling `expansions(t[i], **extra_args)`.)

      Example::

        >>> T = (1, [2, 3])
        >>> list(expansions(T))
        [(1, 2), (1, 3)]

    * If `obj` is a `Template` class instance, then the returned values
      are the result of applying the template to the expansion of each
      of its keywords.

      Example::

        >>> T1 = Template("a=${n}", n=[0,1])
        >>> E = list(expansions(T1))
        >>> len(E)
        2
        >>> Template('a=${n}', n=0) in E
        True
        >>> Template('a=${n}', n=1) in E
        True

      Note that keywords passed to the `expand` invocation override
      the ones used in template construction::

        >>> T2 = Template("a=${n}")
        >>> E = list(expansions(T2, n=[1,3]))
        >>> Template('a=${n}', n=1) in E
        True
        >>> Template('a=${n}', n=3) in E
        True

    * Any other value is returned unchanged.

      Example:

        >>> V = 42
        >>> list(expansions(V))
        [42]

    """
    if isinstance(obj, dict):
        keys = tuple(obj.keys())  # fix a key order
        for items in SetProductIterator(
                *[list(expansions(obj[keys[i]], **extra_args))
                  for i in range(len(keys))]):
            yield dict((keys[i], items[i]) for i in range(len(keys)))
    elif isinstance(obj, tuple):
        for items in SetProductIterator(
                *[list(expansions(u, **extra_args)) for u in obj]):
            yield tuple(items)
    elif isinstance(obj, list):
        for item in obj:
            for result in expansions(item, **extra_args):
                yield result
    elif isinstance(obj, Template):
        for s in obj.expansions(**extra_args):
            yield s
    else:
        yield obj


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="template",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
import sys
import os
import subprocess

from xml.etree import cElementTree as ET
from xml.etree.ElementTree import Element, SubElement, Comment, tostring

SENTIMENTS = [
    'neutral',
    'negative',
    'very negative',
    'positive',
    'very positive']
FIELD = "pdid"
CONTENT = "content"


def Usage():
    print("Usage: wrapper.py <input_file> <output_file>")


def RunParser(input, output):
    if not os.path.isfile(input):
        print("Input file %s not found" % input)
        return 1

    print("Parsing input file... ")

    tree = ET.parse(input)
    root = tree.getroot()

    try:
        fd = open('input_contents.xml', 'w+')

        # Generate simplified version of the input XML file
        for row in list(root):
            field = row.find(FIELD).text
            content = (row.find(CONTENT).text).encode('utf8')
            fd.write("%s\n%s\n" % (field, content))

        fd.close()
    except OSError as osx:
        # Raise and exit
        print("Failed while preparing simplified Content file. Error: %s" % str(osx))
        return 1

    command = '/usr/bin/java -cp "$CORENLP/*" edu.stanford.nlp.sentiment.SentimentPipeline -file input_contents.xml'
    nlp = subprocess.Popen(
        [command],
        shell=True,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE)

    print("Running command %s" % command)
    (stdout, stderr) = nlp.communicate()

    if nlp.returncode != 0:
        print("Execution failed with exit code: %d" % nlp.returncode)
        try:
            res = open(output, 'w+')
            res.write(stderr)
            res.close()
        except Exception as ex:
            print("Failed writing output file %s. Error type %s. Message: %s" % (output, type(ex), str(ex)))
            print("Failed CoreNLP execution with: %s" % stderr)
        return nlp.returncode

    print("Parsing results... ")

    index = 0
    row = None
    sentiment = ""

    for line in [l.strip() for l in stdout.split('\n')]:
        if line.isdigit():
            if row:
                child = SubElement(row, "Sentiment")
                child.tail = "\n"
                child.text = sentiment[1:]
                sentiment = ""
            index = line
            row = root.find("./row/[%s='%s']" % (FIELD, index))
        elif line.lower() in SENTIMENTS:
            sentiment += ",%s" % line

    # Add last sentiment segment to the last ROW element
    child = SubElement(row, "Sentiment")
    child.tail = "\n"
    child.text = sentiment[1:]

    print("Writing results to output file... ")
    tree.write(output)

    print("Done")
    return 0

if __name__ == '__main__':
    if (len(sys.argv) != 3):
        sys.exit(Usage())
    sys.exit(RunParser(sys.argv[1], sys.argv[2]))



#! /usr/bin/env python2.6
#

"""
Script for running over different parameter combinations. 
"""

# Copyright (C) 2011  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>, Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'


from __future__ import absolute_import, print_function
import gc3libs.debug
import re, os
import numpy as np
from supportGc3 import lower, flatten, str2tuple, getIndex, extractVal, str2vals
from supportGc3 import format_newVal, update_parameter_in_file, safe_eval, str2mat, mat2str, getParameter

import shutil

import logbook, sys
from supportGc3 import wrapLogger

class paraLoop(object):
    '''
      Main class implementing a general loop algorithm. Loop over an arbitrary number of variables and values. 
      Additionally the script allows for grouping variables and then specifying a restricted set of parameter
      combinations thereby avoiding to go through all combinations. 
      Should be subclassed together with a SessionBasedScript. 
    '''
    
    
    def __init__(self, verbosity):
        self.logger = wrapLogger(loggerName = __name__ + 'logger', streamVerb = verbosity.upper(), logFile = __name__ + '.log')

    def process_para_file(self, path_to_para_loop):
        """
        Create sets of parameter substituions for the ``forwardPremium``
        input files and yield a unique identifier for each set.  The
        "recipe" for building such sets is given in the ``para.loop``
        file pointed to by argument `path_to_para_loop`.

        In more detail, `process_para_file` is a generator function, that:
         1. parses the `para.loop` file;
         2. computes the distinct sets of substitutions that should be
            applied to ``forwardPremium`` input files; one such set of
            substitutions corresponds to a single ``forwardPremium`` run;
         3. creates and yields a "unique name" string that shall identify
            *this* particular combination of parameters and input files.

        Each generator iteration yields a pair `(jobname,
        substitutions)`, where `jobname` is a unique string and
        `substitutions` is a dictionary mapping file names (as
        specified in the ``paraFile`` column of the ``para.loop``
        file) to quadruples `(var, index, val, regex)`: into each
        input file, the variable `var` (corresponding to the group
        `paraIndex` in `regex`) is given value `val`.

        :param: `path_to_para_loop`: complete pathname of a file in
            ``para.loop`` format.

        """
        params = self._read_para(path_to_para_loop)
        num_params = len(params)
        self.logger.debug("Read %d parameters from file '%s'" % (num_params, path_to_para_loop))

        variables     = params['variables']
        indices       = params['indices']
        paraFiles     = params['paraFiles']
        paraProps     = params['paraProps']
        groups        = np.array(params['groups'], dtype = np.int16)
        groupRestrs   = params['groupRestrs']
        paraFileRegex = params['paraFileRegex']

        # XXX: why not just use `vals = np.copy(params['vals'])` ??
        vals = np.copy(params['vals'])
        #vals = np.empty((num_params,), 'U100')
        #for ixVals, paraVals in enumerate(params['vals']):
            #vals[ixVals] = paraVals

        # remap group numbers so that they start at 0 and increase in
        # steps of 1
        groups = self._remap_groups(groups)
        num_groups = len(np.unique(groups))

        self.logger.debug("Number of groups: %d" % num_groups)
        self.logger.debug('variables: %s' % variables)
        self.logger.debug('indices: %s' % indices)
        self.logger.debug('vals: %s' % vals)
        self.logger.debug('paraFiles: %s' % paraFiles)
        self.logger.debug('groups: %s' % groups)
        self.logger.debug('groupRestrs: %s' % groupRestrs)  
        self.logger.debug('paraFileRegex: %s' % paraFileRegex)

        # check parameter files for consistency
        #
        # NOTE: the weird use of booleans as array indices is a
        # feature of NumPy, see:
        # http://docs.scipy.org/doc/numpy/reference/arrays.indexing.html
        for group in groups:
            groupSelector = (groups == group)
            groupRestriction = np.unique(groupRestrs[groupSelector])
            nGroupRestriction = len(groupRestriction)
            nGroupVariables = sum(groupSelector)
            nSwIndicator = sum(paraProps[groupSelector] == 'swIndicator')
            nRelevGroupVariables = nGroupVariables - nSwIndicator
            if nGroupRestriction != 1:
                raise gc3libs.exceptions.InvalidUsage(
                    "Group restrictions '%s' are inconsistent for group '%s'"
                    % (groupRestriction, group))
            elif nRelevGroupVariables == 1 and groupRestriction[0].lower() == 'lowertr':
                raise gc3libs.exceptions.InvalidUsage(
                    "No sense in using 'lower triangular' restriction"
                    " with just one variable.")

        # Set up groups
        groupBase = []
        groupIndices = []
        metaBase = []
        for ixGroup, group in enumerate(np.unique(groups[groups >= 0])):
            # ixGroup is used as index for groups
            self.logger.debug('At gpremium L200, ixGroup=%s' % ixGroup)
            self.logger.debug('At gpremium L201, group=%s'% group)
            groupBase.append([])
            # Select vars belonging to group 'group'. Leave out switch indicator vars
            # --------------------------------
            relevGroups = (groups == group)
            swVars = (paraProps == 'swIndicator')
            groupSelector = relevGroups
            for ix in range(0, len(relevGroups)):
                if relevGroups[ix] and not swVars[ix]:
                    groupSelector[ix] = True
                else:
                    groupSelector[ix] = False
            #  -------------------------------
            groupRestr = np.unique(groupRestrs[groupSelector])
            self.logger.debug('At gpremium L%d, groupRestr=%s' % (492, groupRestr))
            self.logger.debug('At gpremium L%d, groupSelector=%s' % (493, groupSelector))
            if len(groupRestr) != 1:
                raise gc3libs.exceptions.InvalidUsage(
                    "Groups have different restrictions")
            for groupVals in vals[groupSelector]:
                values = str2vals(groupVals)
                groupBase[group].append(len(np.array(values)))
                self.logger.debug('At gpremium L%d, groupvals=%s' % (500, values))
            groupIndices.append(list(getIndex(groupBase[ixGroup], groupRestr)))
            self.logger.debug('At gpremium L%d, groupIndices=%s' % (503, groupIndices[ixGroup]))
            metaBase.append(len(groupIndices[ixGroup]))

        # Combine groups without restriction
        metaIndices = list(getIndex(metaBase, None))
        nMetaIndices = len(metaIndices)

        self.logger.debug('Summary after establishing groups:')
        self.logger.debug('  groupbase: %s' % groupBase)
        self.logger.debug('  groupindices: %s' % groupIndices)
        self.logger.debug('  metabase: %s' % metaBase)
        self.logger.debug('  metaind: %s' % metaIndices)

        self.logger.debug("Starting enumeration of independent runs...")
        for ixMeta, meta in enumerate(metaIndices):
            self.logger.debug("Loop iteration(ixMeta) %d of %d (%.2f%%)" %
                           (ixMeta+1, nMetaIndices,
                           100.0 * ((1+ixMeta) / nMetaIndices)))

            index = self.getFullIndex(ixMeta, metaIndices,
                                      groupIndices, groups, paraProps, vals)
            self.logger.debug("Index before flattening: %s" % index)
            index = list(flatten(index))
            self.logger.debug('Flattened index: %s' % index)

            for ixVar in range(0, len(variables)):
                self.logger.debug('variable #%d is %s' % (ixVar, variables[ixVar]))

            runDescription = 'para'# os.path.basename(path_to_para_loop)[:-5]
            substs = gc3libs.utils.defaultdict(list)
            for ixVar, var in enumerate(variables):
                self.logger.debug('variable: %s' % variables[ixVar])
                var = variables[ixVar]
                group = groups[ixVar]
                paraFile = paraFiles[ixVar]
                adjustIndex = indices[ixVar]
                val = format_newVal(extractVal(ixVar, vals, index))
                regex = paraFileRegex[ixVar]
                paraIndex = str2tuple(indices[ixVar])
                self.logger.debug('paraIndex: %s' % paraIndex)
                substs[paraFile].append((var, val, paraIndex, regex))
                if (group >= 0) and paraProps[ixVar] != 'swIndicator': 
                    if ixVar < len(variables):
                        runDescription += '_'
                    runDescription += '%s=%s' % (var, val)
            yield (runDescription, substs)



    @staticmethod
    def getFullIndex(ixMeta, metaIndices, groupIndices, groups, paraProps, vals):
        """
        Returns current index of vals for each variable. 
        The current groupIndex is extended with values for special variables. 
        Currently these are: 
          1) group = -1 variables. 
          2) Indicator switch variables. 
        """
        # --- Establish index list ----
        # -----------------------------
        index = []
        for ixGroup, group in enumerate(np.unique(groups)):
            meta = metaIndices[ixMeta]
            groupSelector = (groups == group)
            groupVarIndices = np.where(groups == group)[0]
            if group != -1:
                groupIndex = groupIndices[group][meta[group]]
                curGroupIndex = 0
                extendedGroupIndex = []
                for groupVarIndex in groupVarIndices:
                    paraProp = paraProps[groupVarIndex]
                    if not paraProp == 'swIndicator':
                        extendedGroupIndex.append(groupIndex[curGroupIndex])
                        curGroupIndex += 1
                    else:
                        values = vals[groupVarIndex].split(',')
                        if ixMeta > 0 and groupIndices[group][metaIndices[ixMeta - 1][group]] != groupIndex: # group has changed indices
                            extendedGroupIndex.append(0)
                        elif ixMeta == 0:
                            extendedGroupIndex.append(0)
                        else:
                            extendedGroupIndex.append(1)
                index.append(extendedGroupIndex)      
            else: # group == -1
                groupMinus1 = groups < 0
                if sum(groupMinus1) > 1:
                    self.logger.warning("more than one -1 variable not supported")
                values = vals[groupMinus1][0].split(',')
                nValues = len(values)
                if ixMeta >= nValues:
                    index.append(int(nValues - 1))
                else:
                    index.append(int(ixMeta))
        return index

    def _remap_groups(self, groups):
        """
        Remap group numbers so that:
          1. they start at 0;
          2. there's no gap between two consecutive group numbers.
        """
        group_nrs_map = dict()
        max_group_nr_seen = -1
        for group_nr in np.unique(groups):
            max_group_nr_seen += 1
            group_nrs_map[group_nr] = max_group_nr_seen
        def remap_group_nrs(nr):
            return group_nrs_map[nr]
        return np.apply_along_axis(np.vectorize(remap_group_nrs),
                                   0, groups)

    # format of the `.loop` files
    #  1. comments lines have `#` as first non-whitespace char
    _loop_comment_re = re.compile(r'^ \s* (\#|$)', re.X)
    #  2. columns are separated by 3 or more spaces
    _loop_colsep_re = re.compile(r'\s\s\s+')
    # 3. allowed regexps; each regexp must have exactly 5 groups
    # (named 'a', 'var', 'b', 'val', and 'c'), of which only
    # 'var' and 'val' are signficant; the other ones are copied verbatim.
    _loop_regexps = {
        'bar-separated':(r'([a-z]+[\s\|]+)'
                         r'(\w+)' # variable name
                         r'(\s*[\|]+\s*)' # bars and spaces
                         r'([\w\s\.,;\[\]\-]+)' # value
                         r'(\s*)'),
        'space-separated':(r'(\s*)'
                           r'(\w+)' # variable name
                           r'(\s+)' # spaces (filler)
                           r'([\w\s\.,;\[\]\-]+)' # values
                           r'(\s*)'), # spaces (filler)
    }

    # -- Read para.loop  --
    def _read_para(self, path):
        """
        Read the ``para.loop`` file at `path` and return array of
        parameters.

        Argument `path` can be the full path to a file, which is assumed
        to be in ``para.loop`` syntax, or the path to a directory, in
        which case the the file ``path/para.loop`` is read.
        """
        if os.path.isdir(path):
            path = os.path.join(path, 'para.loop')

        paraLoopFile = open(path, 'r')

        # skip first line (header)
        paraLoopFile.readline()
        
        data = []
        overallMaxFieldLen = 0
        for line in paraLoopFile:
            if self._loop_comment_re.match(line):
                continue
            columns = list(self._loop_colsep_re.split(line.rstrip()))
            columns[7] = self._loop_regexps[columns[7]]
            columns = tuple(columns)
            self.logger.debug("_read_para: got columns: %s" % (str(columns)))
            maxFieldLen = max(map(len, columns))
            if maxFieldLen > overallMaxFieldLen: 
                overallMaxFieldLen = maxFieldLen
#            data.append(np.array(columns, dtype=dt))
            data.append(columns)
            
        paraLoopFile.close()            
        
        # map column name to NumPy type
        fieldFormat = 'U' + str(overallMaxFieldLen)
        dt = {
            'names':[
                'variables',
                'indices',
                'paraFiles',
                'paraProps',
                'groups',
                'groupRestrs',
                'vals',
                'paraFileRegex',
                ],
            'formats':[
                # U for unicode, see http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html
                fieldFormat, # variables
                fieldFormat, # indices
                fieldFormat, # paraFiles
                fieldFormat, # paraProps
                fieldFormat, # groups
                fieldFormat, # groupRestrs
                fieldFormat, # vals
                fieldFormat, # paraFileRegex
                ],
        }    
        params = []
        for columns in data:
            params.append(np.array(columns, dtype=dt))
        params = np.array(params, dtype=dt)


        # by default numpy sorts capital letters with higher priority.
        # Sorting gives me a unique position for a variable in each input vector! 
        ind = np.lexsort((lower(params['paraFiles']),
                          lower(params['variables']),
                          params['groups']),
                         axis = 0)
        return params[ind]

    def writeParaLoop( self, variables, indices = None, paraFiles =  None, paraProps = None, groups = None, groupRestrs = None, paraFileRegex = None, vals = None, desPath = 'para.loopTmp'):
        paraLoopFile = open(desPath, 'w')
        print >> paraLoopFile, '%25s %25s %25s %25s %25s %25s %25s %25s\n' % ('variables', 'Index', 'ParameterFile', 'paraProp', 'Group', 'groupRestrs', 'vals', 'ParaFileRegex')
    
        Nvariables = len(variables)
        if Nvariables == 0: 
            print 'No variables specified'
            sys.exit()
    
        if indices == None:
            indices = [ '(0)' ] * Nvariables
    
        if paraProps == None:
            paraProps = [ 'None' ] * Nvariables
    
        if groups == None:
            groups = [ '0' ] *  Nvariables
    
        if groupRestrs == None:
            groupRestrs = [ 'none' ] * Nvariables 
    
        if paraFileRegex == None:
            paraFileRegex = [ 'bar-separated' ] * Nvariables
    
        if paraFiles == None:
            paraFiles =  [ 'input/parameters.in' ] * Nvariables 
    
    
        for ixVar, variable in enumerate(variables):
        #  print >> paraLoopFile, '%-25s %-25s %-25s %-25s %-25s %-25s %-25s %-25s\n' % (variables[ixVar], indices[ixVar], paraFiles[ixVar], paraProps[ixVar], groups[ixVar], groupRestrs[ixVar], vals[ixVar], paraFileRegex[ixVar])
            print >> paraLoopFile, '%s    %s    %s    %s    %s    %s    %s    %s' % (variables[ixVar], indices[ixVar], paraFiles[ixVar], paraProps[ixVar], groups[ixVar], groupRestrs[ixVar], vals[ixVar], paraFileRegex[ixVar])
    
        return desPath







#!/usr/bin/env python

'''
  Optimization routine for idrisk paper. 
  Uses polynomial to approximate the (curretnly one dim) function and find the root by evaluating a large # of points. 
  Currently the values are normalized with ln. We want to reduce curvature in iBar, our main target variable. This could become a variable normalization function to be more general. 
  Plots can be created to track the optimization progress. 
'''


from __future__ import absolute_import, print_function
import os
import numpy as np
import scipy.optimize
import scipy.interpolate as si
import numpy.polynomial.chebyshev
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab

from pymods.support.support import wrapLogger


np.seterr(all='raise')

class costlyOptimization(object):
  '''
    Simple optimizer with the goal to minimize (expensive) function evaluations. 
    x are para combos
  '''
  def __init__(self, paras):
    self.jobname         = paras['jobname']
    self.xVars           = paras['xVars']
    self.xInitialGuess   = paras['xInitialParaCombo'] # np.array
    self.targetVars      = paras['targetVars']
    self.target_fx       = np.array(paras['target_fx'])
    self.convCrit        = paras['convCrit'] 
    self.converged       = False
    self.makePlots       = paras['plotting']
    self.nUpdates        = 0
    self.optimFolder     = paras['optimFolder']

    self.x               = self.xInitialGuess
    self.fx              = np.array([])

    self.nIntpoints_max  = 4
    self.maxDist         = 10 # maximal distance from second best point
    self.protected_ix    = False
    self.optimRange      = (-1.,0.)
    self.plotRange       = (-1.,0.)

    # Set up logger
    self.logger = wrapLogger(loggerName = 'costlyOptimizationLogger' + self.optimFolder, streamVerb = 'DEBUG', 
                        logFile = os.path.join(self.optimFolder, 'costlyOpt.log'))
    self.logger.debug('initialized new instance of costlyOptimization for jobname = %s' % paras['jobname'])

    
  def updateInterpolationPoints(self, x, fx):
    self.logger.debug('')
    self.logger.debug('entering updateInterpolationPoints')
    self.logger.debug('self.x = \n%s'  % self.x)
    self.logger.debug('self.fx = \n%s' % self.fx)
    self.logger.debug('x = \n%s' % x)
    self.logger.debug('fx = \n%s' % fx)
    if self.nUpdates == 0:
      self.logger.debug('check if fx has more than one value: ')
      self.logger.debug('np.nonzero = %s' % np.nonzero(fx))
      self.logger.debug('np.nonzero(fx)[0] = %s' % np.nonzero(fx)[0])
      numFxVals = len(np.nonzero(fx)[0])
      if not numFxVals > 1:
        self.logger.debug('numFxVals = %s. Sending back failure code. ' % numFxVals)
        return 1
      else: 
        self.logger.debug('numFxVals = %s. Storing values. ' % numFxVals)
      self.logger.debug('storing initial fx values %s corresponding to inital guess \n%s' % (fx, self.xInitialGuess))
      ixParaCombo = 0
      indicesToDelete = []
      for paraCombo,paraComboFx in zip(x, fx):
        # if paraCombo has no paraComboFx return value eliminate it from xInitialGuess
        if not paraComboFx:
          self.logger.debug('adding paraCombo = %s (element = %s) to indicesToDelete because of missing paraComboFx' % (paraCombo, ixParaCombo))
          indicesToDelete.append(ixParaCombo)
        else:
          self.fx = np.append(self.fx, paraComboFx)
        ixParaCombo += 1
      self.logger.debug('self.x before deletion = %s' % self.x)
      self.logger.debug('indicesToDelete = %s' % indicesToDelete)
      self.x = np.delete(self.x, indicesToDelete, axis = 0) # axis = 0 is necessary to keep m*n array
      self.logger.debug('self.x after deletion = %s\n' % self.x)
    else:
      self.logger.debug('check if the only new guess failed. If so terminate to avoid infinite loop')
      if len(fx) == 1:
        if not fx[0]:
          self.logger.debug('Evaluation of new guess at %s failed. Terminating para combo. ' % x)
          return 1
      self.logger.debug('nUpdates > 0.. proceding to append paraCombo')
      for paraCombo,paraComboFx in zip(x, fx):
        if not paraComboFx:
          self.logger.debug('could not evaluate paraCombo = %s because of missing paraComboFx' % paraCombo)
          continue
        self.logger.debug('checking if paracombo %s already exists' % paraCombo)
        if not paraCombo in self.x:
          self.logger.debug('appending para combo %s' % paraCombo)
          self.x  = np.append(self.x, np.array([paraCombo]), 0)
          self.fx = np.append(self.fx, paraComboFx)
        else:
          self.logger.critical('CRITICAL: x = %s already in self.x = %s' % (x, self.x))
    normedDistances = self._computeNormedDistance(self.fx)
    sortedIndicesDistances = np.argsort(normedDistances)
    self.sorted_x = self.x[sortedIndicesDistances]
    self.sorted_fx = self.fx[sortedIndicesDistances]
    self.sortedNormedDistances = normedDistances[sortedIndicesDistances]
    bestIndex = np.argmin(self._computeNormedDistance(self.fx))
    self.best_x = self.x[bestIndex]
    self.best_fx  = self.fx[bestIndex]
    self.logger.debug('done updating interpolation points. self.x = \n%s, \nself.fx = \n%s' % (self.x, self.fx))
    self.logger.debug('best points: best_x = \n%s, \nbest_fx = \n%s' % (self.best_x, self.best_fx))
    nAboveTarget = np.sum(self.sorted_fx - self.target_fx > 0)
    nBelowTarget = np.sum(self.sorted_fx - self.target_fx < 0)
    self.logger.debug('nAboveTarget = %s' % nAboveTarget)
    self.logger.debug('nBelowTarget = %s' % nBelowTarget)
    if nBelowTarget == 0:
      self.logger.critical('nBelowTarget is 0. Abandoning para combo. ')
      return 1
    if nAboveTarget == 0:
      self.logger.critical('nAboveTarget is 0. Abandoning para combo. ')
      return 1
    nAboveTargetInt = np.sum(self.sorted_fx[:self.nIntpoints_max] - self.target_fx > 0)
    nBelowTargetInt = np.sum(self.sorted_fx[:self.nIntpoints_max] - self.target_fx < 0)
    self.logger.debug('nAboveTargetInt = %s' % nAboveTargetInt)
    self.logger.debug('nBelowTargetInt = %s' % nBelowTargetInt)
    if nAboveTargetInt == 0:
      self.logger.debug('nAboveTargetInt == 0. Setting up protected_ix')
      aboveTarget = self.sorted_fx - self.target_fx > 0
      self.protected_ix = np.nonzero(aboveTarget)[0][0] # finds the first occurance of nonzero element in sorted array
    elif nBelowTarget == 0:
      self.logger.debug('nBelowTargetInt == 0. Setting up protected_ix')
      belowTarget = self.sorted_fx - self.target_fx < 0
      self.protected_ix = np.nonzero(belowTarget)[0][0]
    else:
      self.logger.debug('no need to protect any x. setting self.protected_ix = False')
      self.protected_ix = False
    self.logger.debug('protected_ix is = %s' % self.protected_ix)
    self.nUpdates += 1
    if self.nUpdates > 1 and self.makePlots:
      self.plot()
    self.logger.debug('self.x = \n%s'  % self.x)
    self.logger.debug('self.fx = \n%s' % self.fx)
    self.logger.debug('done updateInterpolationPoints')
    self.logger.debug('')
    return 0
    
  def checkConvergence(self):
    self.logger.debug('')
    self.logger.debug('entering checkConvergence...')
    distance = self._computeNormedDistance(self.best_fx)
    if np.all(distance < self.convCrit):
      self.logger.debug('SUCCESS: converged at fx %s with target %s to precision %s' % (self.best_fx, self.target_fx, self.convCrit))
      self.converged = True
    else: 
      self.logger.debug('not converged at fx %s with target %s to precision %s' % (self.best_fx, self.target_fx, self.convCrit))
      self.converged = False
    self.logger.debug('exiting convergence')
    self.logger.debug('')
    return self.converged
    
  def updateApproximation(self):
    self.logger.debug('')
    self.logger.debug('entering updateApproximation')
    self.logger.debug('self.x = \n%s'  % self.x)
    self.logger.debug('self.fx = \n%s' % self.fx)
    if len(self.x[0]) == 1: # we are dealing with the one-dimensional case
      self.logger.debug('self.x[0] == 1')
      xIn = self.sorted_x.copy()
      fIn = self.sorted_fx.copy()
      withinDistance = ( self.sortedNormedDistances[1] / self.sortedNormedDistances ) < 10
      withinDistance[0] = True # need at least two values
      withinDistance[1] = True # 
      if self.protected_ix:
        self.logger.debug('protected exists')
        try: 
          lenProtected_ix = len(self.protected_ix)
        except:
          lenProtected_ix = 1
        actual_nIntpoints_max = self.nIntpoints_max - lenProtected_ix
        withinDistance[self.protected_ix] = True # protected_ix must always be carried over
      else:
        self.logger.debug('protected doesnt exist')
        actual_nIntpoints_max = self.nIntpoints_max
      actual_nIntpoints = np.min([actual_nIntpoints_max, len(xIn)])
      self.logger.debug('actual_nIntpoints = %s' % actual_nIntpoints)
      self.logger.debug('len(xIn) = %s' % len(xIn))
      withinMaxIntpoints = np.array([ True ] * actual_nIntpoints + [ False ] * ( len(xIn) - actual_nIntpoints ) )
      withinMaxIntpoints[self.protected_ix] = True
      self.logger.debug('withinMaxIntpoints = %s' % withinMaxIntpoints)
      self.logger.debug('withinDistance = %s' % withinDistance)
      self.intPoints_ix = withinMaxIntpoints * withinDistance
      xIn = xIn[self.intPoints_ix]
      fIn = fIn[self.intPoints_ix]
      self.logger.debug('after nIntpoints  xIn = %s \n' % xIn)
      self.logger.debug('after nIntpoints  fIn = %s \n' % fIn)
      self.logger.debug('transforming fIn:')
      fIn = np.log((1. - self.target_fx) + fIn)
      self.logger.debug('sending xIn = %s to lagrange\n' % xIn)
      self.logger.debug('sending fIn = %s to lagrange\n' % fIn)
      ### create approximation ###
      self.gx = si.lagrange(xIn[:, 0], fIn)
      ### -------------------- ###
    else:
      self.logger.debug('multi-dimensional case not implemented... exiting')
      os._exit(1)
    if self.makePlots:
      self.plot()
    self.logger.debug('done updating approximation')
    self.logger.debug('')
    
  def generateNewGuess(self):
    def target(x):
      distance = self.gx(x) - self.target_fx
      return distance[0]
    self.logger.debug('')
    self.logger.debug('entering generateNewGuess')
    self.logger.debug('generating new guess for job %s in iteration %s: ' % (self.jobname, self.nUpdates + 1))
    self.logger.debug('current x points: \n%s' % self.x)
    self.logger.debug('current fx points: \n%s' % self.fx)
    if np.all(self.fx - self.target_fx > 0) or np.all(self.fx - self.target_fx < 0):
      self.logger.critical('the initial points %s with fx = \n%s do not contain the zero' % (self.xInitialGuess, self.fx))
    self.logger.debug('Evaluating polynomial to find zero. ')
    xMat = np.linspace(self.optimRange[0], self.optimRange[1], 1.e6)
    fxGrid = np.exp(self.gx(xMat)) - (1. - self.target_fx)
    normedDistances        = self._computeNormedDistance(fxGrid)
    sortedIndicesDistances = np.argsort(normedDistances)
    sortedDistances        = normedDistances[sortedIndicesDistances]
    self.logger.debug('sortedDistances = \n%s' % sortedDistances)
    xMatSorted             = xMat[sortedIndicesDistances]
    fxGridSorted           = fxGrid[sortedIndicesDistances]
    self.logger.debug('xMatSorted = \n%s' % xMatSorted)
    self.logger.debug('fxGridSorted = \n%s' % fxGridSorted)
    topCandidatesBool      = np.abs(sortedDistances - sortedDistances[0]) < 1.e-4
    topCandidates          = np.nonzero(topCandidatesBool)[0]
    if np.sum(topCandidatesBool) > 1: 
      self.logger.debug('More than one zero. Picking the one closest to intial guesses: ')
      xhats = xMatSorted[topCandidatesBool]
      self.logger.debug('Zeros of polynomial are: \n%s' % xhats)
      avgIntpoints = np.mean(self.sorted_x[self.intPoints_ix])
      bestIndex = topCandidates[np.argmin(np.abs(avgIntpoints - xhats))]
    else:
      self.logger.debug('Only one best index. No need to pick one out of many. Easy case...')
      bestIndex = sortedIndicesDistances[0]
    self.logger.debug('bestIndex is %s' % bestIndex)
    xhat = xMatSorted[bestIndex]
    fxhat = fxGridSorted[bestIndex]
    self.logger.debug('sending back optimal value given gx: xhat = %s fxhat = %s' % (xhat, fxhat))
    if xhat in self.x:
      self.logger.critical('xhat = %s already in self.x = %s' % (xhat, self.x))
      os._exit(1)
    self.logger.debug('done generateNewGuess')
    self.logger.debug('')
    return np.array([[xhat]])
  
  def _computeNormedDistance(self, fx):
    fx = np.asanyarray(fx)
    return np.abs(fx - self.target_fx)    

  def plot(self):
    # make plot
    self.logger.debug('making plot')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.grid()
    # plot approximation
    xVals = np.linspace(self.plotRange[0], self.plotRange[1], 1.e3)
    try: 
      fxVals = self.gx(xVals)
    except: 
      self.logger.debug('couldnt create plot. self.gx not defined')
      return
    # plot polynomial
    ax.plot(xVals, fxVals)
    # plot actual values (non standardized)
    ax.plot(self.x, self.fx, marker = 'o', markersize = 2, linestyle = '')
    ax.plot(self.sorted_x[self.intPoints_ix], self.sorted_fx[self.intPoints_ix], marker = 'o', markersize = 2, linestyle = '', color = 'r')
    # xVals = np.linspace(-1., 0., 1.e2)
    # fxVals = np.log((1. - self.target_fx) + fIn)
    figureFile = os.path.join(self.optimFolder, 'lagrangeApprox_' + str(self.nUpdates) + '.eps')
    fig.savefig(figureFile)
    os.system('chmod 660 ' + figureFile)
    self.logger.debug('done making plot')


if __name__ == '__main__':
  a = np.linspace(0.1, 2, 10)
  b = -np.log(a)
  c = si.lagrange(a, b)
  plt.plot(a, c(a))
  plt.plot(a, b)
  #plt.show()
  
  
  fun = lambda x: -np.log(x)
  x   = np.array([0.01, 5])
  
  xhat = np.array([10])
  iter = 0
  while np.abs(fun(xhat)) > 1.e-4:
    print 'iter %d' % iter
    fx  = fun(x)
    gx  = si.lagrange(x, fx)
    x0 = xhat
    xhat = scipy.optimize.newton_krylov(gx, x0)
    print xhat
    x = np.append(x, xhat)
    iter += 1
    
  print 'done'





#! /usr/bin/env python
#
"""
Generic object serialization (using Python's *pickle*/*cPickle* modules).

See the documentation for Python's standard `*pickle* and *cPickle*
modules`__ for more details.

.. __: http://docs.python.org/library/pickle.html

"""
# Copyright (C) 2011-2012, 2018, 2019  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals

from future import standard_library
standard_library.install_aliases()
from builtins import object

import pickle

import sys
PY2 = (sys.version_info[0] == 2)


DEFAULT_PROTOCOL = pickle.HIGHEST_PROTOCOL


class Persistable(object):

    """
    A mix-in class to mark that an object should be persisted by its ID.

    Any instance of this class is saved as an 'external reference'
    when a container holding a reference to it is saved.

    """

    # __slots__ = (
    #     '__weakref__',
    #     'changed',
    #     'persistent_id',
    # )

    def __init__(self, *args, **kwargs):
        # ensure object will be saved next time Store.save() is invoked
        self.changed = True
        if 'persistent_id' in kwargs:
            self.persistent_id = kwargs.pop('persistent_id')
        super(Persistable, self).__init__(*args, **kwargs)

    def __str__(self):
        try:
            return str(self.persistent_id)
        except AttributeError:
            return super(Persistable, self).__str__()

    def __eq__(self, other):
        if id(self) == id(other):
            return True
        try:
            return self.persistent_id == other.persistent_id
        except AttributeError:
            # fall back to Python object comparison
            return super(Persistable, self) == other

    def __ne__(self, other):
        return not self.__eq__(other)


def make_pickler(driver, stream, root, protocol=pickle.HIGHEST_PROTOCOL):
    return _PicklerWithPersistentID(driver, root, stream, protocol=protocol)


def make_unpickler(driver, stream):
    return _UnpicklerWithPersistentID(driver, stream)


class _PicklerWithPersistentID(pickle.Pickler):

    """Used internally to provide `persistent_id` support to *cPickle*.

    This class is needed because:

    * we want to save each `Persistable`:class: object as a separate record,
    * we want to use the *cPickle* module for performance reasons.

    Check the `documentation of Python's *pickle* module`__ for
    details on `persistent_id` support:

    .. __: http://goo.gl/CCknrT

    """

    def __init__(self, driver, root, stream, **kwargs):
        self._root = root
        self._driver = driver
        self._stream = stream
        pickle.Pickler.__init__(self, stream, **kwargs)

    def persistent_id(self, obj):
        if obj is self._root:
            return None
        elif isinstance(obj, Persistable):
            if (not hasattr(obj, 'persistent_id')
                or getattr(obj, 'changed', True)):
                self._driver.save(obj)
            return obj.persistent_id


class _UnpicklerWithPersistentID(pickle.Unpickler):

    """Used internally to provide `persistent_id` support to *cPickle*.

    This class is needed because:

    * we want to save each `Persistable`:class: object as a separate record,
    * we want to use the *cPickle* module for performance reasons.

    Check the `documentation of Python's *pickle* module`__ for
    details on the differences between *pickle* and *cPickle* modules.

    .. __: http://goo.gl/CCknrT

    """

    def __init__(self, driver, *args, **kwargs):
        self._driver = driver
        pickle.Unpickler.__init__(self, *args, **kwargs)

    def persistent_load(self, id_):
        return self._driver.load(id_)

    # Override pickle's default `load_build()` to use `setattr()` upon
    # `AttributeError` and not just on `RuntimeError`.  This works only on
    # Python2, since Python3 has a completely different implemenation of `pickle`.
    if PY2:
        def load_build(self):
            stack = self.stack
            state = stack.pop()
            inst = stack[-1]
            setstate = getattr(inst, "__setstate__", None)
            if setstate:
                setstate(state)
                return
            slotstate = None
            if isinstance(state, tuple) and len(state) == 2:
                state, slotstate = state
            if state:
                try:
                    d = inst.__dict__
                    try:
                        for k, v in state.iteritems():
                            d[intern(k)] = v
                    # keys in state don't have to be strings
                    # don't blow up, but don't go out of our way
                    except TypeError:
                        d.update(state)

                except (AttributeError, RuntimeError):
                    # XXX In restricted execution, the instance's __dict__
                    # is not accessible.  Use the old way of unpickling
                    # the instance variables.  This is a semantic
                    # difference when unpickling in restricted
                    # vs. unrestricted modes.
                    # Note, however, that cPickle has never tried to do the
                    # .update() business, and always uses
                    #     PyObject_SetItem(inst.__dict__, key, value) in a
                    # loop over state.items().
                    for k, v in state.items():
                        setattr(inst, k, v)
            if slotstate:
                for k, v in slotstate.items():
                    setattr(inst, k, v)

if PY2:
    # the `dispatch` table is a class attribute of `pickle.Unpickler` so
    # we need to take a copy, otherwise overriding the handler for
    # `pickle.BUILD` with a method bound to class
    # `_UnpicklerWithPersistentID` makes it fail when called from an
    # instance of `pickle.Unpickler`
    _UnpicklerWithPersistentID.dispatch = pickle.Unpickler.dispatch.copy()

    # register the modified `load_build` as handler for `pickle.BUILD`
    _UnpicklerWithPersistentID.dispatch[pickle.BUILD] = _UnpicklerWithPersistentID.load_build



#! /usr/bin/env python
#
#   gsearchlight.py -- Front-end script for evaluating cosmo searchlight.
#
#   Copyright (C) 2016, 2017  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `Matlab` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gsearchlight.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

__version__ = 'development version (SVN $Revision$)'
# summary of user-visible changes
__changelog__ = """
  2016-12-01:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: http://code.google.com/p/gc3pie/issues/detail?id=95
if __name__ == "__main__":
    import gsearchlight
    gsearchlight.GsearchlightScript().run()

import os
import sys
import time
import tempfile

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds

DEFAULT_ITERATIONS=1000
DEFAULT_CHUNKS=1
DEFAULT_REMOTE_OUTPUT_FILE="result"
MATLAB_CMD="matlab -nosplash -nodisplay -nodesktop -r \'{main_function}({iterations} {mask} {fn} {output});quit\'"

## custom application class
class GsearchlightApplication(Application):
    """
    Custom class to wrap the execution of the Matlab function
    over a subset of the total number of events.
    """
    application_name = 'gsearchlight'

    def __init__(self, iterations, mask, mask_hdr, fn_file, matlab_file, **extra_args):

        executables = []
        inputs = dict()
        outputs = dict()

        inputs[matlab_file] = os.path.basename(matlab_file)
        matlab_function = inputs[matlab_file].split('.')[0]
        inputs[mask] = os.path.basename(mask)
        inputs[mask_hdr] = os.path.basename(mask_hdr)
        inputs[fn_file] = os.path.basename(fn_file)

        # arguments = MATLAB_CMD.format(main_function=matlab_function,
        #                               iterations=iterations,
        #                               mask=inputs[mask_file],
        #                               fn=inputs[fn_file],
        #                               output=DEFAULT_REMOTE_OUTPUT_FILE)

        wrapper = resource_filename(Requirement.parse("gc3pie"),
                                    "gc3libs/etc/run_matlab.sh")
        inputs[wrapper] = "./wrapper.sh"

        arguments = "./wrapper.sh %s %d %s %s %s" % (matlab_function,
                                                     iterations,
                                                     inputs[mask],
                                                     inputs[fn_file],
                                                     DEFAULT_REMOTE_OUTPUT_FILE)

        # Set output
        outputs[DEFAULT_REMOTE_OUTPUT_FILE] = DEFAULT_REMOTE_OUTPUT_FILE

        gc3libs.log.debug("Creating application for executing: %s",
                          arguments)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gsearchlight.log',
            join=True,
            executables = executables,
            **extra_args)


class GsearchlightScript(SessionBasedScript):
    """
    Take total number of events and create a list of chunked events.
    For each chunk, run the provided Matlab function.

    The ``gsearchlight`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gsearchlight`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gsearchlight``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GsearchlightApplication,
            stats_only_for = GsearchlightApplication,
            )

    def setup_args(self):

        self.add_param('Mfunct',
                       type=existing_file,
                       help="Full path to Matlab function to execute.")

        self.add_param('mask',
                       type=existing_file,
                       help="Full path to Mask file.")

        self.add_param('fn',
                       type=existing_file,
                       help="Full path to Fn file.")

    def setup_options(self):
        self.add_param("-I", "--iterations", metavar="INT",
                       type=positive_int,
                       dest="niter", default=DEFAULT_ITERATIONS,
                       help="How to split the edges input data set. "
                       "Default: %(default)s.")

        self.add_param("-G", "--group", metavar="INT",
                       type=positive_int,
                       dest="chunk", default=DEFAULT_CHUNKS,
                       help="Group events together and process them in groups. "
                       "Default: %(default)s.")

    def parse_args(self):
        """
        check that 'mask' file is provided with its .hdr complement
        """

        mask_filename = self.params.mask.split(".")[0]
        assert os.path.isfile(mask_filename+'.hdr'), \
            "Mask file %s.hdr missing" % mask_filename
        self.params.maskhdr = mask_filename+".hdr"

    def new_tasks(self, extra):
        """
        Read content of 'command_file'
        For each command line, generate a new Application
        """
        tasks = []

        for iteration in get_events(self.params.niter,self.params.chunk):
            extra_args = extra.copy()
            tasks.append(GsearchlightApplication(
                iteration,
                self.params.mask,
                self.params.maskhdr,
                self.params.fn,
                self.params.Mfunct,
                **extra_args))

        return tasks

def get_events(niter,chunk):
    iterations = []
    for i in range(0,(niter-(niter%chunk)),chunk):
        iterations.append(chunk)
    last = niter%chunk
    if last > 0:
        iterations.append(last)
    return iterations
